@@ -1,6 +1,7 @@
 /*
     ccid.c: CCID common code
     Copyright (C) 2003-2005   Ludovic Rousseau
+    Copyright (C) 2009-2012   Advanced Card Systems Ltd.
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -33,9 +34,11 @@
 #include "defs.h"
 #include "ccid_ifdhandler.h"
 #include "commands.h"
+#include "acr38cmd.h"
 #include "ccid_usb.h"
 
-static int GetFirmwareVersion(unsigned int reader_index);
+static int ACR83_GetFirmwareVersion(unsigned int reader_index, unsigned int *pFirmwareVersion);
+static int ACR1222_GetFirmwareVersion(unsigned int reader_index, char *firmwareVersion, unsigned int *pFirmwareVersionLen);
 
 /*****************************************************************************
  *
@@ -45,7 +48,7 @@ static int GetFirmwareVersion(unsigned int reader_index);
 int ccid_open_hack_pre(unsigned int reader_index)
 {
 	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
-	int doInterruptRead = 1;
+	int doInterruptRead = 0;	// Disable InterruptRead
 	int i;
 
 	switch (ccid_descriptor->readerID)
@@ -71,11 +74,9 @@ int ccid_open_hack_pre(unsigned int reader_index)
 			doInterruptRead = 0;
 			break;
 
-		default:
-			// Disable InterruptRead for all ACS CCID readers
-			// ACR88U and ACR128U does not timeout on Mac OS X
-			doInterruptRead = 0;
-
+		case ACS_ACR122U:
+		case ACS_AET62_PICC_READER:
+		case ACS_AET62_1SAM_PICC_READER:
 			// Enable polling mode (ACR122 v2.06)
 			for (i = 0; i < 10; i++)
 			{
@@ -86,6 +87,9 @@ int ccid_open_hack_pre(unsigned int reader_index)
 				}
 			}
 			break;
+
+		default:
+			break;
 	}
 
 	/* CCID */
@@ -364,59 +368,237 @@ int ccid_open_hack_post(unsigned int reader_index)
 			break;
 #endif
 
+		case ACS_ACR33U_A1_3SAM_ICC_READER:
+		case ACS_ACR33U_A2_3SAM_ICC_READER:
+		case ACS_ACR33U_A3_3SAM_ICC_READER:
+		case ACS_ACR33U_4SAM_ICC_READER:
+			if (ccid_descriptor->bCurrentSlotIndex > 1)
+				ccid_descriptor->isSamSlot = 1;	// SAM
+			break;
+
 		case ACS_ACR83U:
 			{
-				unsigned char command[] = { 0x04, 0x00, 0x00, 0x00, 0x00 };
-				unsigned int commandLen = sizeof(command);
-				unsigned char response[3 + 6];
-				unsigned int responseLen = sizeof(response);
 				unsigned int firmwareVersion;
 
 				ccid_descriptor->wLcdLayout = 0x0210;
-				if (CmdEscape(reader_index, command, commandLen, response, &responseLen) == IFD_SUCCESS)
+
+				DEBUG_INFO("Getting ACR83U firmware version...");
+				if (ACR83_GetFirmwareVersion(reader_index, &firmwareVersion))
 				{
-					if ((responseLen >= 7) && (response[0] == 0x84))
+					DEBUG_INFO2("ACR83U firmware version: 0x%04X", firmwareVersion);
+					if (firmwareVersion >= 0x4500)
 					{
-						firmwareVersion = (response[5] << 8) | response[6];
-						DEBUG_INFO2("ACR83U firmware version: 0x%04X", firmwareVersion);
-						if (firmwareVersion >= 0x4500)
-						{
-							ccid_descriptor->dwFeatures &= ~CCID_CLASS_EXCHANGE_MASK;
-							ccid_descriptor->dwFeatures |= CCID_CLASS_SHORT_APDU;
-						}
-						else
-						{
-							// Disable PIN support (firmware version < 0x4500)
-							ccid_descriptor->bPINSupport = 0;
-						}
+						// ACR83U uses short APDU exchange
+						ccid_descriptor->dwFeatures &= ~CCID_CLASS_EXCHANGE_MASK;
+						ccid_descriptor->dwFeatures |= CCID_CLASS_SHORT_APDU;
+					}
+					else
+					{
+						// Disable PIN support (firmware version < 0x4500)
+						ccid_descriptor->bPINSupport = 0;
 					}
 				}
 			}
 			break;
 
 		case ACS_APG8201:
+		case ACS_APG8201Z:
 			ccid_descriptor->wLcdLayout = 0x0210;
-		case ACS_ACR85_PINPAD_READER_ICC:
+
 			// APG8201 uses short APDU exchange
 			ccid_descriptor->dwFeatures &= ~CCID_CLASS_EXCHANGE_MASK;
 			ccid_descriptor->dwFeatures |= CCID_CLASS_SHORT_APDU;
 			break;
 
-		// Enable PICC
-		case ACS_ACR1222_DUAL_READER:
-		case ACS_ACR1222_1SAM_DUAL_READER:
-			DEBUG_INFO("Getting firmware version...");
-			ccid_descriptor->firmwareVersion = GetFirmwareVersion(reader_index);
-			DEBUG_INFO2("ACR1222: %d", ccid_descriptor->firmwareVersion);
-			if (ccid_descriptor->bCurrentSlotIndex == 1)
+		case ACS_ACR85_PINPAD_READER_ICC:
 			{
-				DEBUG_INFO("Enabling PICC...");
-				EnablePicc(reader_index, 1);
+				unsigned int firmwareVersion;
+
+				// ACR85 ICC uses short APDU exchange
+				ccid_descriptor->dwFeatures &= ~CCID_CLASS_EXCHANGE_MASK;
+				ccid_descriptor->dwFeatures |= CCID_CLASS_SHORT_APDU;
+
+				DEBUG_INFO("Getting ACR85 ICC firmware version...");
+				if (ACR83_GetFirmwareVersion(reader_index, &firmwareVersion))
+				{
+					DEBUG_INFO2("ACR85 ICC firmware version: 0x%04X", firmwareVersion);
+
+					// Set firmware fix enabled
+					ccid_descriptor->firmwareFixEnabled = (firmwareVersion == 0x0001);
+					DEBUG_INFO2("Firmware fix enabled: %d", ccid_descriptor->firmwareFixEnabled);
+				}
 			}
 			break;
+
+		case ACS_ACR1222_1SAM_PICC_READER:
+			if (ccid_descriptor->bCurrentSlotIndex == 1)
+				ccid_descriptor->isSamSlot = 1;	// SAM
+			break;
+
+		case ACS_ACR1222_DUAL_READER:
+		case ACS_ACR1222_1SAM_DUAL_READER:
+			{
+				char firmwareVersion[30];
+				unsigned int firmwareVersionLen = sizeof(firmwareVersion);
+
+				DEBUG_INFO("Getting ACR1222 firmware version...");
+				if (ACR1222_GetFirmwareVersion(reader_index, firmwareVersion, &firmwareVersionLen))
+				{
+					DEBUG_INFO2("ACR1222 firmware version: %s", firmwareVersion);
+
+					ccid_descriptor->firmwareFixEnabled = (strcmp(firmwareVersion, "ACR1222U_V401") == 0);
+					DEBUG_INFO2("Firmware fix enabled: %d", ccid_descriptor->firmwareFixEnabled);
+
+					if ((ccid_descriptor->firmwareFixEnabled) &&
+						(ccid_descriptor->bCurrentSlotIndex == 1))
+					{
+						DEBUG_INFO("Enabling PICC...");
+						EnablePicc(reader_index, 1);
+					}
+				}
+
+				if (ccid_descriptor->bCurrentSlotIndex == 2)
+					ccid_descriptor->isSamSlot = 1;	// SAM
+			}
+			break;
+
+		case ACS_ACR1222_3S_PICC_READER:
+			if (ccid_descriptor->bCurrentSlotIndex > 0)
+				ccid_descriptor->isSamSlot = 1;	// SAM
+			break;
+
 		case ACS_ACR85_PINPAD_READER_PICC:
-			DEBUG_INFO("Enabling PICC...");
-			EnablePicc(reader_index, 1);
+			{
+				char firmwareVersion[30];
+				unsigned int firmwareVersionLen = sizeof(firmwareVersion);
+
+				DEBUG_INFO("Getting ACR85 PICC firmware version...");
+				if (ACR1222_GetFirmwareVersion(reader_index, firmwareVersion, &firmwareVersionLen))
+				{
+					DEBUG_INFO2("ACR85 PICC firmware version: %s", firmwareVersion);
+
+					ccid_descriptor->firmwareFixEnabled = (strcmp(firmwareVersion, "ACR1222U_V402") == 0);
+					DEBUG_INFO2("Firmware fix enabled: %d", ccid_descriptor->firmwareFixEnabled);
+
+					if (ccid_descriptor->firmwareFixEnabled)
+					{
+						DEBUG_INFO("Enabling PICC...");
+						EnablePicc(reader_index, 1);
+					}
+				}
+			}
+			break;
+
+		case ACS_ACR38U:
+		case ACS_ACR38U_SAM:
+		case IRIS_SCR21U:
+		case ACS_AET65_1SAM_ICC_READER:
+			if (ccid_descriptor->bCurrentSlotIndex == 0)
+			{
+				char firmwareVersion[11];	// Firmware version
+
+				// Get firmware version
+				memset(firmwareVersion, 0, sizeof(firmwareVersion));
+				if (ACR38_GetFirmwareVersion(reader_index, firmwareVersion) == IFD_SUCCESS)
+				{
+					DEBUG_INFO2("Firmware: %s", firmwareVersion);
+
+					if (ccid_descriptor->readerID == IRIS_SCR21U)
+					{
+						// Adjust maximum data rate
+						if (strcmp(firmwareVersion, "ACR38-1042") == 0)
+							ccid_descriptor->dwMaxDataRate = 43010;	// MCU
+					}
+				}
+			}
+			else
+			{
+				// Adjust maximum data rate
+				ccid_descriptor->dwMaxDataRate = 10752;	// SAM
+				ccid_descriptor->isSamSlot = 1;
+			}
+			break;
+
+		case ACS_ACR88U:
+			// Adjust maximum data rate
+			if (ccid_descriptor->bCurrentSlotIndex < 2)
+				ccid_descriptor->dwMaxDataRate = 116129;	// MCU
+			else
+			{
+				ccid_descriptor->dwMaxDataRate = 9677;		// SAM
+				ccid_descriptor->isSamSlot = 1;
+			}
+			break;
+
+		case ACS_ACR89_ICC_READER:
+		case ACS_ACR89_FP_READER:
+			if (ccid_descriptor->bCurrentSlotIndex > 1)
+				ccid_descriptor->isSamSlot = 1;	// SAM
+			break;
+
+		case ACS_ACR89_DUAL_READER:
+			if (ccid_descriptor->bCurrentSlotIndex > 2)
+				ccid_descriptor->isSamSlot = 1;	// SAM
+			break;
+
+		case ACS_ACR128U:
+			// Adjust features and maximum data rate
+			if (ccid_descriptor->bCurrentSlotIndex == 0)
+			{
+				ccid_descriptor->dwFeatures = 0x000204BA;	// MCU
+				ccid_descriptor->dwMaxDataRate = 116129;
+			}
+			else if (ccid_descriptor->bCurrentSlotIndex == 1)
+			{
+				ccid_descriptor->dwFeatures = 0x0002047A;	// Contactless
+				ccid_descriptor->dwMaxDataRate = 116129;
+			}
+			else
+			{
+				ccid_descriptor->dwFeatures = 0x000204BA;	// SAM
+				ccid_descriptor->dwMaxDataRate = 9677;
+				ccid_descriptor->isSamSlot = 1;
+			}
+			break;
+
+		case ACS_ACR1281_1S_DUAL_READER:
+			// Adjust features and maximum data rate
+			if (ccid_descriptor->bCurrentSlotIndex == 0)
+			{
+				ccid_descriptor->dwFeatures = 0x000204BA;	// MCU
+				ccid_descriptor->dwMaxDataRate = 344100;
+			}
+			else if (ccid_descriptor->bCurrentSlotIndex == 1)
+			{
+				ccid_descriptor->dwFeatures = 0x0004047A;	// Contactless
+				ccid_descriptor->dwMaxDataRate = 344100;
+			}
+			else
+			{
+				ccid_descriptor->dwFeatures = 0x000204BA;	// SAM
+				ccid_descriptor->dwMaxDataRate = 125000;
+				ccid_descriptor->isSamSlot = 1;
+			}
+			break;
+
+		case ACS_ACR1281_2S_CL_READER:
+			// Adjust features and maximum data rate
+			if (ccid_descriptor->bCurrentSlotIndex == 1)
+			{
+				ccid_descriptor->dwFeatures = 0x0004047A;	// Contactless
+				ccid_descriptor->dwMaxDataRate = 344100;
+			}
+			else
+			{
+				ccid_descriptor->dwFeatures = 0x000204BA;	// SAM
+				ccid_descriptor->dwMaxDataRate = 125000;
+				ccid_descriptor->isSamSlot = 1;
+			}
+			break;
+
+		case ACS_ACR1281_1S_PICC_READER:
+			if (ccid_descriptor->bCurrentSlotIndex == 1)
+				ccid_descriptor->isSamSlot = 1;	// SAM
 			break;
 
 		default:
@@ -540,21 +722,84 @@ void ccid_error(int error, const char *file, int line, const char *function)
 
 		default:
 			if ((error >= 1) && (error <= 127))
-			{
 				(void)snprintf(var_text, sizeof(var_text), "error on byte %d",
 					error);
-				text = var_text;
-			}
 			else
 				(void)snprintf(var_text, sizeof(var_text),
 					"Unknown CCID error: 0x%02X", error);
-				text = var_text;
+
+			text = var_text;
 			break;
 	}
 	log_msg(PCSC_LOG_ERROR, "%s:%d:%s %s", file, line, function, text);
 
 } /* ccid_error */
 
+void acr38_error(int error, const char *file, int line, const char *function)
+{
+	const char *text;
+	char var_text[30];
+
+	switch (error)
+	{
+		case 0x00:
+			text = "Success";
+			break;
+
+		case 0xF4:
+			text = "Procedure byte conflict";
+			break;
+
+		case 0xF6:
+			text = "Bad length";
+			break;
+
+		case 0xF7:
+			text = "Bad Fi/Di";
+			break;
+
+		case 0xF8:
+			text = "Bad ATR TS";
+			break;
+
+		case 0xF9:
+			text = "ICC not powered up";
+			break;
+
+		case 0xFA:
+			text = "ICC not inserted";
+			break;
+
+		case 0xFB:
+			text = "Hardware error";
+			break;
+
+		case 0xFC:
+			text = "XFE overrun";
+			break;
+
+		case 0xFD:
+			text = "XFE parity error";
+			break;
+
+		case 0xFE:
+			text = "ICC mute";
+			break;
+
+		case 0xFF:
+			text = "Command aborted";
+			break;
+
+		default:
+			(void)snprintf(var_text, sizeof(var_text),
+				"Unknown ACR38 error: 0x%02X", error);
+			text = var_text;
+			break;
+	}
+
+	log_msg(PCSC_LOG_ERROR, "%s:%d:%s %s", file, line, function, text);
+}
+
 // Enable PICC
 void EnablePicc(unsigned int reader_index, int enabled)
 {
@@ -601,21 +846,43 @@ void EnablePicc(unsigned int reader_index, int enabled)
 	}
 }
 
-// Get firmware version (ACR1222)
-static int GetFirmwareVersion(unsigned int reader_index)
+static int ACR83_GetFirmwareVersion(unsigned int reader_index, unsigned int *pFirmwareVersion)
 {
-	int firmwareVersion = -1;
+	int ret = 0;
+	unsigned char command[] = { 0x04, 0x00, 0x00, 0x00, 0x00 };
+	unsigned int commandLen = sizeof(command);
+	unsigned char response[3 + 6];
+	unsigned int responseLen = sizeof(response);
 
-	unsigned char getFirmwareVersion[] = { 0xE0, 0x00, 0x00, 0x18, 0x00 };
-	unsigned char response[300];
-	int responseLen;
-
-	responseLen = sizeof(response);
-	if (CmdEscape(reader_index, getFirmwareVersion, sizeof(getFirmwareVersion), response, &responseLen) == IFD_SUCCESS)
+	if (CmdEscape(reader_index, command, commandLen, response, &responseLen) == IFD_SUCCESS)
 	{
-		if (responseLen > 5)
-			sscanf((char *) (response + 5), "ACR1222U_V%d", &firmwareVersion);
+		if ((responseLen >= 7) && (response[0] == 0x84))
+		{
+			*pFirmwareVersion = (response[5] << 8) | response[6];
+			ret = 1;
+		}
 	}
 
-	return firmwareVersion;
+	return ret;
+}
+
+static int ACR1222_GetFirmwareVersion(unsigned int reader_index, char *firmwareVersion, unsigned int *pFirmwareVersionLen)
+{
+	int ret = 0;
+	unsigned char command[] = { 0xE0, 0x00, 0x00, 0x18, 0x00 };
+	unsigned char response[300];
+	unsigned int responseLen = sizeof(response);
+
+	if (CmdEscape(reader_index, command, sizeof(command), response, &responseLen) == IFD_SUCCESS)
+	{
+		if (*pFirmwareVersionLen >= responseLen - 5 + 1)
+		{
+			*pFirmwareVersionLen = responseLen - 5;
+			memcpy(firmwareVersion, response + 5, *pFirmwareVersionLen);
+			firmwareVersion[*pFirmwareVersionLen] = '\0';
+			ret = 1;
+		}
+	}
+
+	return ret;
 }
