@@ -1,6 +1,7 @@
 /*
     ccid_usb.c: USB access routines using the libusb library
     Copyright (C) 2003-2009   Ludovic Rousseau
+    Copyright (C) 2009-2012   Advanced Card Systems Ltd.
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -103,7 +104,7 @@ static int get_end_points(struct usb_device *dev, _usbDevice *usbdevice, int num
 int ccid_check_firmware(struct usb_device *dev);
 static unsigned int *get_data_rates(unsigned int reader_index,
 	struct usb_device *dev, int num);
-	
+
 #ifdef __APPLE__
 static void *CardDetectionThread(void *pParam);
 #endif
@@ -399,7 +400,8 @@ status_t OpenUSBByName(unsigned int reader_index, /*@null@*/ char *device)
 						}
 						// Simulate ACR1281 Dual Reader (composite device) as multi-slot reader
 						else if ((ACS_ACR1281_DUAL_READER_QPBOC == readerID) ||
-							(ACS_ACR1281_DUAL_READER_BSI == readerID))
+							(ACS_ACR1281_DUAL_READER_BSI == readerID) ||
+							(ACS_ACR1281_1S_PICC_READER == readerID))
 						{
 							// the CCID interfaces are 0 and 1
 							interface_number = static_interface - 1;
@@ -535,8 +537,16 @@ again:
 
 					if (usb_interface->altsetting->extralen != 54)
 					{
-						// ACR88U and ACR128U does not have altsetting
-						if ((readerID != ACS_ACR88U) && (readerID != ACS_ACR128U))
+						// These readers do not have altsetting
+						if ((readerID != ACS_ACR38U) &&
+							(readerID != ACS_ACR38U_SAM) &&
+							(readerID != IRIS_SCR21U) &&
+							(readerID != ACS_AET65_1SAM_ICC_READER) &&
+							(readerID != ACS_CRYPTOMATE) &&
+							(readerID != ACS_ACR88U) &&
+							(readerID != ACS_ACR128U) &&
+							(readerID != ACS_ACR1281_1S_DUAL_READER) &&
+							(readerID != ACS_ACR1281_2S_CL_READER))
 						{
 							(void)usb_close(dev_handle);
 							DEBUG_CRITICAL4("Extra field for %s/%s has a wrong length: %d", bus->dirname, dev->filename, usb_interface->altsetting->extralen);
@@ -610,8 +620,45 @@ again:
 					// Store bcdDevice for firmware version checking
 					usbDevice[reader_index].ccid.bcdDevice = dev->descriptor.bcdDevice;
 
-					// ACR88U and ACR128U does not have altsetting
-					if (readerID == ACS_ACR88U)
+					// These readers do not have altsetting
+					if ((readerID == ACS_ACR38U) || (readerID == ACS_CRYPTOMATE))
+					{
+						usbDevice[reader_index].ccid.dwFeatures = 0x00010030;
+						usbDevice[reader_index].ccid.wLcdLayout = 0;
+						usbDevice[reader_index].ccid.bPINSupport = 0;
+						usbDevice[reader_index].ccid.dwMaxCCIDMessageLength = 271;
+						usbDevice[reader_index].ccid.dwMaxIFSD = 248;
+						usbDevice[reader_index].ccid.dwDefaultClock = 4000;
+						usbDevice[reader_index].ccid.dwMaxDataRate = 229390;
+						usbDevice[reader_index].ccid.bMaxSlotIndex = 0;
+						usbDevice[reader_index].ccid.bCurrentSlotIndex = 0;
+						usbDevice[reader_index].ccid.readTimeout = DEFAULT_COM_READ_TIMEOUT;
+						usbDevice[reader_index].ccid.arrayOfSupportedDataRates = NULL;
+						usbDevice[reader_index].ccid.bInterfaceProtocol = PROTOCOL_ACR38;
+						usbDevice[reader_index].ccid.bNumEndpoints = 3;
+						usbDevice[reader_index].ccid.dwSlotStatus = IFD_ICC_PRESENT;
+						usbDevice[reader_index].ccid.bVoltageSupport = 0x07;
+					}
+					else if ((readerID == ACS_ACR38U_SAM) || (readerID == IRIS_SCR21U) ||
+						(readerID == ACS_AET65_1SAM_ICC_READER))
+					{
+						usbDevice[reader_index].ccid.dwFeatures = 0x00010030;
+						usbDevice[reader_index].ccid.wLcdLayout = 0;
+						usbDevice[reader_index].ccid.bPINSupport = 0;
+						usbDevice[reader_index].ccid.dwMaxCCIDMessageLength = 271;
+						usbDevice[reader_index].ccid.dwMaxIFSD = 248;
+						usbDevice[reader_index].ccid.dwDefaultClock = 4000;
+						usbDevice[reader_index].ccid.dwMaxDataRate = 229390;
+						usbDevice[reader_index].ccid.bMaxSlotIndex = 1;
+						usbDevice[reader_index].ccid.bCurrentSlotIndex = 0;
+						usbDevice[reader_index].ccid.readTimeout = DEFAULT_COM_READ_TIMEOUT;
+						usbDevice[reader_index].ccid.arrayOfSupportedDataRates = NULL;
+						usbDevice[reader_index].ccid.bInterfaceProtocol = PROTOCOL_ACR38;
+						usbDevice[reader_index].ccid.bNumEndpoints = 3;
+						usbDevice[reader_index].ccid.dwSlotStatus = IFD_ICC_PRESENT;
+						usbDevice[reader_index].ccid.bVoltageSupport = 0x07;
+					}
+					else if (readerID == ACS_ACR88U)
 					{
 						usbDevice[reader_index].ccid.dwFeatures = 0x000204BA;
 						usbDevice[reader_index].ccid.wLcdLayout = 0x0815;
@@ -619,7 +666,7 @@ again:
 						usbDevice[reader_index].ccid.dwMaxCCIDMessageLength = 271;
 						usbDevice[reader_index].ccid.dwMaxIFSD = 248;
 						usbDevice[reader_index].ccid.dwDefaultClock = 3600;
-						usbDevice[reader_index].ccid.dwMaxDataRate = 115200;
+						usbDevice[reader_index].ccid.dwMaxDataRate = 116129;
 						usbDevice[reader_index].ccid.bMaxSlotIndex = 4;
 						usbDevice[reader_index].ccid.bCurrentSlotIndex = 0;
 						usbDevice[reader_index].ccid.readTimeout = DEFAULT_COM_READ_TIMEOUT;
@@ -637,7 +684,26 @@ again:
 						usbDevice[reader_index].ccid.dwMaxCCIDMessageLength = 271;
 						usbDevice[reader_index].ccid.dwMaxIFSD = 248;
 						usbDevice[reader_index].ccid.dwDefaultClock = 3600;
-						usbDevice[reader_index].ccid.dwMaxDataRate = 115200;
+						usbDevice[reader_index].ccid.dwMaxDataRate = 116129;
+						usbDevice[reader_index].ccid.bMaxSlotIndex = 2;
+						usbDevice[reader_index].ccid.bCurrentSlotIndex = 0;
+						usbDevice[reader_index].ccid.readTimeout = DEFAULT_COM_READ_TIMEOUT;
+						usbDevice[reader_index].ccid.arrayOfSupportedDataRates = NULL;
+						usbDevice[reader_index].ccid.bInterfaceProtocol = 0;
+						usbDevice[reader_index].ccid.bNumEndpoints = 3;
+						usbDevice[reader_index].ccid.dwSlotStatus = IFD_ICC_PRESENT;
+						usbDevice[reader_index].ccid.bVoltageSupport = 0x03;
+					}
+					else if ((readerID == ACS_ACR1281_1S_DUAL_READER) ||
+						(readerID == ACS_ACR1281_2S_CL_READER))
+					{
+						usbDevice[reader_index].ccid.dwFeatures = 0x000204BA;
+						usbDevice[reader_index].ccid.wLcdLayout = 0;
+						usbDevice[reader_index].ccid.bPINSupport = 0;
+						usbDevice[reader_index].ccid.dwMaxCCIDMessageLength = 271;
+						usbDevice[reader_index].ccid.dwMaxIFSD = 248;
+						usbDevice[reader_index].ccid.dwDefaultClock = 4000;
+						usbDevice[reader_index].ccid.dwMaxDataRate = 344100;
 						usbDevice[reader_index].ccid.bMaxSlotIndex = 2;
 						usbDevice[reader_index].ccid.bCurrentSlotIndex = 0;
 						usbDevice[reader_index].ccid.readTimeout = DEFAULT_COM_READ_TIMEOUT;
@@ -660,7 +726,8 @@ again:
 						usbDevice[reader_index].ccid.dwMaxDataRate = dw2i(usb_interface->altsetting->extra, 23);
 
 						// Fix ACR1222 incorrect max slot index
-						if (readerID == ACS_ACR1222_DUAL_READER)
+						if ((readerID == ACS_ACR1222_1SAM_PICC_READER) ||
+							(readerID == ACS_ACR1222_DUAL_READER))
 							usbDevice[reader_index].ccid.bMaxSlotIndex = 1;
 						else if (readerID == ACS_ACR1222_1SAM_DUAL_READER)
 							usbDevice[reader_index].ccid.bMaxSlotIndex = 2;
@@ -694,8 +761,8 @@ again:
 					// Initialize array of bStatus
 					memset(usbDevice[reader_index].ccid.bStatus, 0xFF, numSlots * sizeof(unsigned char));
 
-					// Initialize firmware version (ACR1222)
-					usbDevice[reader_index].ccid.firmwareVersion = -1;
+					// Initialize firmware fix enabled
+					usbDevice[reader_index].ccid.firmwareFixEnabled = FALSE;
 
 					// Simulate ACR85 as multi-slot reader
 					if (readerID != ACS_ACR85_PINPAD_READER_PICC)
@@ -708,12 +775,21 @@ again:
 						usbDevice[reader_index].ccid.piccReaderIndex = -1;
 						usbDevice[reader_index].ccid.pPiccReaderIndex = &usbDevice[reader_index].ccid.piccReaderIndex;
 					}
+
+					// Initialize card voltage (ACR38U, ACR38U-SAM and SCR21U)
+					usbDevice[reader_index].ccid.cardVoltage = 0;
+
+					// Initialize card type (ACR38U, ACR38U-SAM and SCR21U)
+					usbDevice[reader_index].ccid.cardType = 0;
+
+					// Initialize isSamSlot
+					usbDevice[reader_index].ccid.isSamSlot = FALSE;
 #ifdef __APPLE__
 					// Initialize terminated flag to false
 					usbDevice[reader_index].terminated = FALSE;
 					usbDevice[reader_index].pTerminated = &usbDevice[reader_index].terminated;
 					usbDevice[reader_index].ccid.pbStatusLock = &usbDevice[reader_index].ccid.bStatusLock;
-					
+
 					// Create bStatus lock
 					ret = pthread_mutex_init(usbDevice[reader_index].ccid.pbStatusLock, NULL);
 					if (ret != 0)
@@ -735,7 +811,6 @@ again:
 						return STATUS_UNSUCCESSFUL;
 					}
 #endif
-					DEBUG_INFO2("bMaxSlotIndex: %d", usbDevice[reader_index].ccid.bMaxSlotIndex);
 					goto end;
 				}
 			}
@@ -833,43 +908,120 @@ status_t ReadUSB(unsigned int reader_index, unsigned int * length,
 	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
 	int duplicate_frame = 0;
 
-read_again:
+	unsigned char epBuffer[64];
+	int responseReceived;
+	int headerReceived;
+	unsigned int bufferLen;
+	unsigned int readLen;
+	unsigned int dataLen;
+	unsigned int tempLen;
+
 	(void)snprintf(debug_header, sizeof(debug_header), "<- %06X ",
 		(int)reader_index);
 
-	rv = usb_bulk_read(usbDevice[reader_index].handle,
-		usbDevice[reader_index].bulk_in, (char *)buffer, *length,
-		usbDevice[reader_index].ccid.readTimeout * 1000);
-
-	if (rv < 0)
+	if (ccid_descriptor->bInterfaceProtocol == PROTOCOL_ACR38)
 	{
-		*length = 0;
-		DEBUG_CRITICAL4("usb_bulk_read(%s/%s): %s",
-			usbDevice[reader_index].dirname, usbDevice[reader_index].filename,
-			strerror(errno));
+		responseReceived = FALSE;
+		headerReceived = FALSE;
+		bufferLen = *length;
+		readLen = 0;
+		dataLen = 0;
+		tempLen = 0;
 
-		if (ENODEV == errno)
-			return STATUS_NO_SUCH_DEVICE;
-
-		return STATUS_UNSUCCESSFUL;
-	}
-
-	*length = rv;
-
-	DEBUG_XXD(debug_header, buffer, *length);
-
-#define BSEQ_OFFSET 6
-	if ((*length >= BSEQ_OFFSET)
-		&& (buffer[BSEQ_OFFSET] < *ccid_descriptor->pbSeq -1))
-	{
-		duplicate_frame++;
-		if (duplicate_frame > 10)
+		while (!responseReceived)
 		{
-			DEBUG_CRITICAL("Too many duplicate frame detected");
+			// Read data from BULK IN endpoint
+			rv = usb_bulk_read(usbDevice[reader_index].handle,
+				usbDevice[reader_index].bulk_in, (char *) epBuffer, sizeof(epBuffer),
+				usbDevice[reader_index].ccid.readTimeout * 1000);
+			if (rv < 0)
+			{
+				*length = 0;
+				DEBUG_CRITICAL4("usb_bulk_read(%s/%s): %s",
+					usbDevice[reader_index].dirname, usbDevice[reader_index].filename,
+					strerror(errno));
+
+				if (ENODEV == errno)
+					return STATUS_NO_SUCH_DEVICE;
+
+				return STATUS_UNSUCCESSFUL;
+			}
+
+			DEBUG_XXD(debug_header, epBuffer, rv);
+
+			// Copy data to buffer
+			if (readLen + rv <= bufferLen)
+				memcpy(buffer + readLen, epBuffer, rv);
+			readLen += rv;
+
+			if (!headerReceived)
+			{
+				if ((readLen >= 4) &&
+					(buffer[0] == 0x01))	// Header
+				{
+					// Header is received
+					headerReceived = TRUE;
+
+					// Get data length
+					dataLen = ((buffer[2] << 8) | buffer[3]) + 4;
+				}
+			}
+
+			if (headerReceived)
+			{
+				if (readLen >= dataLen)
+				{
+					// Response is received
+					*length = readLen;
+					responseReceived = TRUE;
+				}
+			}
+		}
+
+		// Insufficient buffer
+		if (readLen > bufferLen)
+		{
+			DEBUG_CRITICAL("Insufficient buffer");
 			return STATUS_UNSUCCESSFUL;
 		}
-		DEBUG_INFO("Duplicate frame detected");
-		goto read_again;
+	}
+	else
+	{
+	read_again:
+		rv = usb_bulk_read(usbDevice[reader_index].handle,
+			usbDevice[reader_index].bulk_in, (char *)buffer, *length,
+			usbDevice[reader_index].ccid.readTimeout * 1000);
+
+		if (rv < 0)
+		{
+			*length = 0;
+			DEBUG_CRITICAL4("usb_bulk_read(%s/%s): %s",
+				usbDevice[reader_index].dirname, usbDevice[reader_index].filename,
+				strerror(errno));
+
+			if (ENODEV == errno)
+				return STATUS_NO_SUCH_DEVICE;
+
+			return STATUS_UNSUCCESSFUL;
+		}
+
+		*length = rv;
+
+		DEBUG_XXD(debug_header, buffer, *length);
+
+#define BSEQ_OFFSET 6
+		if ((*length >= BSEQ_OFFSET)
+			&& (buffer[BSEQ_OFFSET] < *ccid_descriptor->pbSeq -1))
+		{
+			duplicate_frame++;
+			if (duplicate_frame > 10)
+			{
+				DEBUG_CRITICAL("Too many duplicate frame detected");
+				return STATUS_UNSUCCESSFUL;
+			}
+			DEBUG_INFO("Duplicate frame detected");
+			goto read_again;
+		}
 	}
 
 	return STATUS_SUCCESS;
